"""Generated wrapper for TWRegistry Solidity contract."""

# pylint: disable=too-many-arguments

import json
from typing import (  # pylint: disable=unused-import
    Any,
    List,
    Optional,
    Tuple,
    Union,
)

from eth_utils import to_checksum_address
from mypy_extensions import TypedDict  # pylint: disable=unused-import
from hexbytes import HexBytes
from web3 import Web3
from web3.contract import ContractFunction
from web3.datastructures import AttributeDict
from web3.providers.base import BaseProvider

from zero_ex.contract_wrappers.bases import ContractMethod, Validator
from zero_ex.contract_wrappers.tx_params import TxParams


# Try to import a custom validator class definition; if there isn't one,
# declare one that we can instantiate for the default argument to the
# constructor for TWRegistry below.
try:
    # both mypy and pylint complain about what we're doing here, but this
    # works just fine, so their messages have been disabled here.
    from . import (  # type: ignore # pylint: disable=import-self
        TWRegistryValidator,
    )
except ImportError:

    class TWRegistryValidator(Validator):  # type: ignore
        """No-op input validator."""


try:
    from .middleware import MIDDLEWARE  # type: ignore
except ImportError:
    pass


class DefaultAdminRoleMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the DEFAULT_ADMIN_ROLE method."""

    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        contract_function: ContractFunction,
    ):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address)
        self._underlying_method = contract_function

    def call(self, tx_params: Optional[TxParams] = None) -> Union[bytes, str]:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        tx_params = super().normalize_tx_params(tx_params)
        returned = self._underlying_method().call(tx_params.as_dict())
        return Union[bytes, str](returned)

    def send_transaction(
        self, tx_params: Optional[TxParams] = None
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().transact(tx_params.as_dict())

    def build_transaction(self, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().buildTransaction(tx_params.as_dict())

    def estimate_gas(self, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().estimateGas(tx_params.as_dict())


class OperatorRoleMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the OPERATOR_ROLE method."""

    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        contract_function: ContractFunction,
    ):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address)
        self._underlying_method = contract_function

    def call(self, tx_params: Optional[TxParams] = None) -> Union[bytes, str]:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        tx_params = super().normalize_tx_params(tx_params)
        returned = self._underlying_method().call(tx_params.as_dict())
        return returned

    def send_transaction(
        self, tx_params: Optional[TxParams] = None
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().transact(tx_params.as_dict())

    def build_transaction(self, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().buildTransaction(tx_params.as_dict())

    def estimate_gas(self, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().estimateGas(tx_params.as_dict())


class AddMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the add method."""

    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, deployer: str, deployment: str):
        """Validate the inputs to the add method."""
        self.validator.assert_valid(
            method_name="add",
            parameter_name="_deployer",
            argument_value=deployer,
        )
        deployer = self.validate_and_checksum_address(deployer)
        self.validator.assert_valid(
            method_name="add",
            parameter_name="_deployment",
            argument_value=deployment,
        )
        deployment = self.validate_and_checksum_address(deployment)
        return (deployer, deployment)

    def call(
        self, deployer: str, deployment: str, tx_params: Optional[TxParams] = None
    ) -> None:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (deployer, deployment) = self.validate_and_normalize_inputs(
            deployer, deployment
        )
        tx_params = super().normalize_tx_params(tx_params)
        self._underlying_method(deployer, deployment).call(tx_params.as_dict())

    def send_transaction(
        self, deployer: str, deployment: str, tx_params: Optional[TxParams] = None
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (deployer, deployment) = self.validate_and_normalize_inputs(
            deployer, deployment
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(deployer, deployment).transact(
            tx_params.as_dict()
        )

    def build_transaction(
        self, deployer: str, deployment: str, tx_params: Optional[TxParams] = None
    ) -> dict:
        """Construct calldata to be used as input to the method."""
        (deployer, deployment) = self.validate_and_normalize_inputs(
            deployer, deployment
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(deployer, deployment).buildTransaction(
            tx_params.as_dict()
        )

    def estimate_gas(
        self, deployer: str, deployment: str, tx_params: Optional[TxParams] = None
    ) -> int:
        """Estimate gas consumption of method call."""
        (deployer, deployment) = self.validate_and_normalize_inputs(
            deployer, deployment
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(deployer, deployment).estimateGas(
            tx_params.as_dict()
        )


class CountMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the count method."""

    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, deployer: str):
        """Validate the inputs to the count method."""
        self.validator.assert_valid(
            method_name="count",
            parameter_name="_deployer",
            argument_value=deployer,
        )
        deployer = self.validate_and_checksum_address(deployer)
        return deployer

    def call(self, deployer: str, tx_params: Optional[TxParams] = None) -> int:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (deployer) = self.validate_and_normalize_inputs(deployer)
        tx_params = super().normalize_tx_params(tx_params)
        returned = self._underlying_method(deployer).call(tx_params.as_dict())
        return int(returned)

    def send_transaction(
        self, deployer: str, tx_params: Optional[TxParams] = None
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (deployer) = self.validate_and_normalize_inputs(deployer)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(deployer).transact(tx_params.as_dict())

    def build_transaction(
        self, deployer: str, tx_params: Optional[TxParams] = None
    ) -> dict:
        """Construct calldata to be used as input to the method."""
        (deployer) = self.validate_and_normalize_inputs(deployer)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(deployer).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, deployer: str, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (deployer) = self.validate_and_normalize_inputs(deployer)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(deployer).estimateGas(tx_params.as_dict())


class GetAllMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the getAll method."""

    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, deployer: str):
        """Validate the inputs to the getAll method."""
        self.validator.assert_valid(
            method_name="getAll",
            parameter_name="_deployer",
            argument_value=deployer,
        )
        deployer = self.validate_and_checksum_address(deployer)
        return deployer

    def call(self, deployer: str, tx_params: Optional[TxParams] = None) -> List[str]:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (deployer) = self.validate_and_normalize_inputs(deployer)
        tx_params = super().normalize_tx_params(tx_params)
        returned = self._underlying_method(deployer).call(tx_params.as_dict())
        return [str(element) for element in returned]

    def send_transaction(
        self, deployer: str, tx_params: Optional[TxParams] = None
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (deployer) = self.validate_and_normalize_inputs(deployer)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(deployer).transact(tx_params.as_dict())

    def build_transaction(
        self, deployer: str, tx_params: Optional[TxParams] = None
    ) -> dict:
        """Construct calldata to be used as input to the method."""
        (deployer) = self.validate_and_normalize_inputs(deployer)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(deployer).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, deployer: str, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (deployer) = self.validate_and_normalize_inputs(deployer)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(deployer).estimateGas(tx_params.as_dict())


class GetRoleAdminMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the getRoleAdmin method."""

    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, role: Union[bytes, str]):
        """Validate the inputs to the getRoleAdmin method."""
        self.validator.assert_valid(
            method_name="getRoleAdmin",
            parameter_name="role",
            argument_value=role,
        )
        return role

    def call(
        self, role: Union[bytes, str], tx_params: Optional[TxParams] = None
    ) -> Union[bytes, str]:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (role) = self.validate_and_normalize_inputs(role)
        tx_params = super().normalize_tx_params(tx_params)
        returned = self._underlying_method(role).call(tx_params.as_dict())
        return Union[bytes, str](returned)

    def send_transaction(
        self, role: Union[bytes, str], tx_params: Optional[TxParams] = None
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (role) = self.validate_and_normalize_inputs(role)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role).transact(tx_params.as_dict())

    def build_transaction(
        self, role: Union[bytes, str], tx_params: Optional[TxParams] = None
    ) -> dict:
        """Construct calldata to be used as input to the method."""
        (role) = self.validate_and_normalize_inputs(role)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role).buildTransaction(tx_params.as_dict())

    def estimate_gas(
        self, role: Union[bytes, str], tx_params: Optional[TxParams] = None
    ) -> int:
        """Estimate gas consumption of method call."""
        (role) = self.validate_and_normalize_inputs(role)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role).estimateGas(tx_params.as_dict())


class GetRoleMemberMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the getRoleMember method."""

    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, role: Union[bytes, str], index: int):
        """Validate the inputs to the getRoleMember method."""
        self.validator.assert_valid(
            method_name="getRoleMember",
            parameter_name="role",
            argument_value=role,
        )
        self.validator.assert_valid(
            method_name="getRoleMember",
            parameter_name="index",
            argument_value=index,
        )
        # safeguard against fractional inputs
        index = int(index)
        return (role, index)

    def call(
        self, role: Union[bytes, str], index: int, tx_params: Optional[TxParams] = None
    ) -> str:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (role, index) = self.validate_and_normalize_inputs(role, index)
        tx_params = super().normalize_tx_params(tx_params)
        returned = self._underlying_method(role, index).call(tx_params.as_dict())
        return str(returned)

    def send_transaction(
        self, role: Union[bytes, str], index: int, tx_params: Optional[TxParams] = None
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (role, index) = self.validate_and_normalize_inputs(role, index)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, index).transact(tx_params.as_dict())

    def build_transaction(
        self, role: Union[bytes, str], index: int, tx_params: Optional[TxParams] = None
    ) -> dict:
        """Construct calldata to be used as input to the method."""
        (role, index) = self.validate_and_normalize_inputs(role, index)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, index).buildTransaction(
            tx_params.as_dict()
        )

    def estimate_gas(
        self, role: Union[bytes, str], index: int, tx_params: Optional[TxParams] = None
    ) -> int:
        """Estimate gas consumption of method call."""
        (role, index) = self.validate_and_normalize_inputs(role, index)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, index).estimateGas(tx_params.as_dict())


class GetRoleMemberCountMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the getRoleMemberCount method."""

    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, role: Union[bytes, str]):
        """Validate the inputs to the getRoleMemberCount method."""
        self.validator.assert_valid(
            method_name="getRoleMemberCount",
            parameter_name="role",
            argument_value=role,
        )
        return role

    def call(
        self, role: Union[bytes, str], tx_params: Optional[TxParams] = None
    ) -> int:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (role) = self.validate_and_normalize_inputs(role)
        tx_params = super().normalize_tx_params(tx_params)
        returned = self._underlying_method(role).call(tx_params.as_dict())
        return int(returned)

    def send_transaction(
        self, role: Union[bytes, str], tx_params: Optional[TxParams] = None
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (role) = self.validate_and_normalize_inputs(role)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role).transact(tx_params.as_dict())

    def build_transaction(
        self, role: Union[bytes, str], tx_params: Optional[TxParams] = None
    ) -> dict:
        """Construct calldata to be used as input to the method."""
        (role) = self.validate_and_normalize_inputs(role)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role).buildTransaction(tx_params.as_dict())

    def estimate_gas(
        self, role: Union[bytes, str], tx_params: Optional[TxParams] = None
    ) -> int:
        """Estimate gas consumption of method call."""
        (role) = self.validate_and_normalize_inputs(role)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role).estimateGas(tx_params.as_dict())


class GrantRoleMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the grantRole method."""

    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, role: Union[bytes, str], account: str):
        """Validate the inputs to the grantRole method."""
        self.validator.assert_valid(
            method_name="grantRole",
            parameter_name="role",
            argument_value=role,
        )
        self.validator.assert_valid(
            method_name="grantRole",
            parameter_name="account",
            argument_value=account,
        )
        account = self.validate_and_checksum_address(account)
        return (role, account)

    def call(
        self,
        role: Union[bytes, str],
        account: str,
        tx_params: Optional[TxParams] = None,
    ) -> None:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        self._underlying_method(role, account).call(tx_params.as_dict())

    def send_transaction(
        self,
        role: Union[bytes, str],
        account: str,
        tx_params: Optional[TxParams] = None,
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, account).transact(tx_params.as_dict())

    def build_transaction(
        self,
        role: Union[bytes, str],
        account: str,
        tx_params: Optional[TxParams] = None,
    ) -> dict:
        """Construct calldata to be used as input to the method."""
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, account).buildTransaction(
            tx_params.as_dict()
        )

    def estimate_gas(
        self,
        role: Union[bytes, str],
        account: str,
        tx_params: Optional[TxParams] = None,
    ) -> int:
        """Estimate gas consumption of method call."""
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, account).estimateGas(tx_params.as_dict())


class HasRoleMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the hasRole method."""

    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, role: Union[bytes, str], account: str):
        """Validate the inputs to the hasRole method."""
        self.validator.assert_valid(
            method_name="hasRole",
            parameter_name="role",
            argument_value=role,
        )
        self.validator.assert_valid(
            method_name="hasRole",
            parameter_name="account",
            argument_value=account,
        )
        account = self.validate_and_checksum_address(account)
        return (role, account)

    def call(
        self,
        role: Union[bytes, str],
        account: str,
        tx_params: Optional[TxParams] = None,
    ) -> bool:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        returned = self._underlying_method(role, account).call(tx_params.as_dict())
        return bool(returned)

    def send_transaction(
        self,
        role: Union[bytes, str],
        account: str,
        tx_params: Optional[TxParams] = None,
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, account).transact(tx_params.as_dict())

    def build_transaction(
        self,
        role: Union[bytes, str],
        account: str,
        tx_params: Optional[TxParams] = None,
    ) -> dict:
        """Construct calldata to be used as input to the method."""
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, account).buildTransaction(
            tx_params.as_dict()
        )

    def estimate_gas(
        self,
        role: Union[bytes, str],
        account: str,
        tx_params: Optional[TxParams] = None,
    ) -> int:
        """Estimate gas consumption of method call."""
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, account).estimateGas(tx_params.as_dict())


class IsTrustedForwarderMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the isTrustedForwarder method."""

    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, forwarder: str):
        """Validate the inputs to the isTrustedForwarder method."""
        self.validator.assert_valid(
            method_name="isTrustedForwarder",
            parameter_name="forwarder",
            argument_value=forwarder,
        )
        forwarder = self.validate_and_checksum_address(forwarder)
        return forwarder

    def call(self, forwarder: str, tx_params: Optional[TxParams] = None) -> bool:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (forwarder) = self.validate_and_normalize_inputs(forwarder)
        tx_params = super().normalize_tx_params(tx_params)
        returned = self._underlying_method(forwarder).call(tx_params.as_dict())
        return bool(returned)

    def send_transaction(
        self, forwarder: str, tx_params: Optional[TxParams] = None
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (forwarder) = self.validate_and_normalize_inputs(forwarder)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(forwarder).transact(tx_params.as_dict())

    def build_transaction(
        self, forwarder: str, tx_params: Optional[TxParams] = None
    ) -> dict:
        """Construct calldata to be used as input to the method."""
        (forwarder) = self.validate_and_normalize_inputs(forwarder)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(forwarder).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, forwarder: str, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (forwarder) = self.validate_and_normalize_inputs(forwarder)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(forwarder).estimateGas(tx_params.as_dict())


class MulticallMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the multicall method."""

    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, data: List[Union[bytes, str]]):
        """Validate the inputs to the multicall method."""
        self.validator.assert_valid(
            method_name="multicall",
            parameter_name="data",
            argument_value=data,
        )
        return data

    def call(
        self, data: List[Union[bytes, str]], tx_params: Optional[TxParams] = None
    ) -> List[Union[bytes, str]]:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (data) = self.validate_and_normalize_inputs(data)
        tx_params = super().normalize_tx_params(tx_params)
        returned = self._underlying_method(data).call(tx_params.as_dict())
        return [Union[bytes, str](element) for element in returned]

    def send_transaction(
        self, data: List[Union[bytes, str]], tx_params: Optional[TxParams] = None
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (data) = self.validate_and_normalize_inputs(data)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(data).transact(tx_params.as_dict())

    def build_transaction(
        self, data: List[Union[bytes, str]], tx_params: Optional[TxParams] = None
    ) -> dict:
        """Construct calldata to be used as input to the method."""
        (data) = self.validate_and_normalize_inputs(data)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(data).buildTransaction(tx_params.as_dict())

    def estimate_gas(
        self, data: List[Union[bytes, str]], tx_params: Optional[TxParams] = None
    ) -> int:
        """Estimate gas consumption of method call."""
        (data) = self.validate_and_normalize_inputs(data)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(data).estimateGas(tx_params.as_dict())


class RemoveMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the remove method."""

    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, deployer: str, deployment: str):
        """Validate the inputs to the remove method."""
        self.validator.assert_valid(
            method_name="remove",
            parameter_name="_deployer",
            argument_value=deployer,
        )
        deployer = self.validate_and_checksum_address(deployer)
        self.validator.assert_valid(
            method_name="remove",
            parameter_name="_deployment",
            argument_value=deployment,
        )
        deployment = self.validate_and_checksum_address(deployment)
        return (deployer, deployment)

    def call(
        self, deployer: str, deployment: str, tx_params: Optional[TxParams] = None
    ) -> None:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (deployer, deployment) = self.validate_and_normalize_inputs(
            deployer, deployment
        )
        tx_params = super().normalize_tx_params(tx_params)
        self._underlying_method(deployer, deployment).call(tx_params.as_dict())

    def send_transaction(
        self, deployer: str, deployment: str, tx_params: Optional[TxParams] = None
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (deployer, deployment) = self.validate_and_normalize_inputs(
            deployer, deployment
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(deployer, deployment).transact(
            tx_params.as_dict()
        )

    def build_transaction(
        self, deployer: str, deployment: str, tx_params: Optional[TxParams] = None
    ) -> dict:
        """Construct calldata to be used as input to the method."""
        (deployer, deployment) = self.validate_and_normalize_inputs(
            deployer, deployment
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(deployer, deployment).buildTransaction(
            tx_params.as_dict()
        )

    def estimate_gas(
        self, deployer: str, deployment: str, tx_params: Optional[TxParams] = None
    ) -> int:
        """Estimate gas consumption of method call."""
        (deployer, deployment) = self.validate_and_normalize_inputs(
            deployer, deployment
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(deployer, deployment).estimateGas(
            tx_params.as_dict()
        )


class RenounceRoleMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the renounceRole method."""

    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, role: Union[bytes, str], account: str):
        """Validate the inputs to the renounceRole method."""
        self.validator.assert_valid(
            method_name="renounceRole",
            parameter_name="role",
            argument_value=role,
        )
        self.validator.assert_valid(
            method_name="renounceRole",
            parameter_name="account",
            argument_value=account,
        )
        account = self.validate_and_checksum_address(account)
        return (role, account)

    def call(
        self,
        role: Union[bytes, str],
        account: str,
        tx_params: Optional[TxParams] = None,
    ) -> None:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        self._underlying_method(role, account).call(tx_params.as_dict())

    def send_transaction(
        self,
        role: Union[bytes, str],
        account: str,
        tx_params: Optional[TxParams] = None,
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, account).transact(tx_params.as_dict())

    def build_transaction(
        self,
        role: Union[bytes, str],
        account: str,
        tx_params: Optional[TxParams] = None,
    ) -> dict:
        """Construct calldata to be used as input to the method."""
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, account).buildTransaction(
            tx_params.as_dict()
        )

    def estimate_gas(
        self,
        role: Union[bytes, str],
        account: str,
        tx_params: Optional[TxParams] = None,
    ) -> int:
        """Estimate gas consumption of method call."""
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, account).estimateGas(tx_params.as_dict())


class RevokeRoleMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the revokeRole method."""

    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, role: Union[bytes, str], account: str):
        """Validate the inputs to the revokeRole method."""
        self.validator.assert_valid(
            method_name="revokeRole",
            parameter_name="role",
            argument_value=role,
        )
        self.validator.assert_valid(
            method_name="revokeRole",
            parameter_name="account",
            argument_value=account,
        )
        account = self.validate_and_checksum_address(account)
        return (role, account)

    def call(
        self,
        role: Union[bytes, str],
        account: str,
        tx_params: Optional[TxParams] = None,
    ) -> None:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        self._underlying_method(role, account).call(tx_params.as_dict())

    def send_transaction(
        self,
        role: Union[bytes, str],
        account: str,
        tx_params: Optional[TxParams] = None,
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, account).transact(tx_params.as_dict())

    def build_transaction(
        self,
        role: Union[bytes, str],
        account: str,
        tx_params: Optional[TxParams] = None,
    ) -> dict:
        """Construct calldata to be used as input to the method."""
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, account).buildTransaction(
            tx_params.as_dict()
        )

    def estimate_gas(
        self,
        role: Union[bytes, str],
        account: str,
        tx_params: Optional[TxParams] = None,
    ) -> int:
        """Estimate gas consumption of method call."""
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, account).estimateGas(tx_params.as_dict())


class SupportsInterfaceMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the supportsInterface method."""

    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, interface_id: Union[bytes, str]):
        """Validate the inputs to the supportsInterface method."""
        self.validator.assert_valid(
            method_name="supportsInterface",
            parameter_name="interfaceId",
            argument_value=interface_id,
        )
        return interface_id

    def call(
        self, interface_id: Union[bytes, str], tx_params: Optional[TxParams] = None
    ) -> bool:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (interface_id) = self.validate_and_normalize_inputs(interface_id)
        tx_params = super().normalize_tx_params(tx_params)
        returned = self._underlying_method(interface_id).call(tx_params.as_dict())
        return bool(returned)

    def send_transaction(
        self, interface_id: Union[bytes, str], tx_params: Optional[TxParams] = None
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (interface_id) = self.validate_and_normalize_inputs(interface_id)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(interface_id).transact(tx_params.as_dict())

    def build_transaction(
        self, interface_id: Union[bytes, str], tx_params: Optional[TxParams] = None
    ) -> dict:
        """Construct calldata to be used as input to the method."""
        (interface_id) = self.validate_and_normalize_inputs(interface_id)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(interface_id).buildTransaction(
            tx_params.as_dict()
        )

    def estimate_gas(
        self, interface_id: Union[bytes, str], tx_params: Optional[TxParams] = None
    ) -> int:
        """Estimate gas consumption of method call."""
        (interface_id) = self.validate_and_normalize_inputs(interface_id)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(interface_id).estimateGas(tx_params.as_dict())


# pylint: disable=too-many-public-methods,too-many-instance-attributes
class TWRegistry:
    """Wrapper class for TWRegistry Solidity contract."""

    default_admin_role: DefaultAdminRoleMethod
    """Constructor-initialized instance of
    :class:`DefaultAdminRoleMethod`.
    """

    operator_role: OperatorRoleMethod
    """Constructor-initialized instance of
    :class:`OperatorRoleMethod`.
    """

    add: AddMethod
    """Constructor-initialized instance of
    :class:`AddMethod`.
    """

    count: CountMethod
    """Constructor-initialized instance of
    :class:`CountMethod`.
    """

    get_all: GetAllMethod
    """Constructor-initialized instance of
    :class:`GetAllMethod`.
    """

    get_role_admin: GetRoleAdminMethod
    """Constructor-initialized instance of
    :class:`GetRoleAdminMethod`.
    """

    get_role_member: GetRoleMemberMethod
    """Constructor-initialized instance of
    :class:`GetRoleMemberMethod`.
    """

    get_role_member_count: GetRoleMemberCountMethod
    """Constructor-initialized instance of
    :class:`GetRoleMemberCountMethod`.
    """

    grant_role: GrantRoleMethod
    """Constructor-initialized instance of
    :class:`GrantRoleMethod`.
    """

    has_role: HasRoleMethod
    """Constructor-initialized instance of
    :class:`HasRoleMethod`.
    """

    is_trusted_forwarder: IsTrustedForwarderMethod
    """Constructor-initialized instance of
    :class:`IsTrustedForwarderMethod`.
    """

    multicall: MulticallMethod
    """Constructor-initialized instance of
    :class:`MulticallMethod`.
    """

    remove: RemoveMethod
    """Constructor-initialized instance of
    :class:`RemoveMethod`.
    """

    renounce_role: RenounceRoleMethod
    """Constructor-initialized instance of
    :class:`RenounceRoleMethod`.
    """

    revoke_role: RevokeRoleMethod
    """Constructor-initialized instance of
    :class:`RevokeRoleMethod`.
    """

    supports_interface: SupportsInterfaceMethod
    """Constructor-initialized instance of
    :class:`SupportsInterfaceMethod`.
    """

    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        validator: TWRegistryValidator = None,
    ):
        """Get an instance of wrapper for smart contract.

        :param web3_or_provider: Either an instance of `web3.Web3`:code: or
            `web3.providers.base.BaseProvider`:code:
        :param contract_address: where the contract has been deployed
        :param validator: for validation of method inputs.
        """
        # pylint: disable=too-many-statements

        self.contract_address = contract_address

        if not validator:
            validator = TWRegistryValidator(web3_or_provider, contract_address)

        web3 = None
        if isinstance(web3_or_provider, BaseProvider):
            web3 = Web3(web3_or_provider)
        elif isinstance(web3_or_provider, Web3):
            web3 = web3_or_provider
        else:
            raise TypeError(
                "Expected parameter 'web3_or_provider' to be an instance of either"
                + " Web3 or BaseProvider"
            )

        # if any middleware was imported, inject it
        try:
            MIDDLEWARE
        except NameError:
            pass
        else:
            try:
                for middleware in MIDDLEWARE:
                    web3.middleware_onion.inject(
                        middleware["function"],
                        layer=middleware["layer"],
                    )
            except ValueError as value_error:
                if value_error.args == (
                    "You can't add the same un-named instance twice",
                ):
                    pass

        self._web3_eth = web3.eth

        functions = self._web3_eth.contract(
            address=to_checksum_address(contract_address), abi=TWRegistry.abi()
        ).functions

        self.default_admin_role = DefaultAdminRoleMethod(
            web3_or_provider, contract_address, functions.DEFAULT_ADMIN_ROLE
        )

        self.operator_role = OperatorRoleMethod(
            web3_or_provider, contract_address, functions.OPERATOR_ROLE
        )

        self.add = AddMethod(
            web3_or_provider, contract_address, functions.add, validator
        )

        self.count = CountMethod(
            web3_or_provider, contract_address, functions.count, validator
        )

        self.get_all = GetAllMethod(
            web3_or_provider, contract_address, functions.getAll, validator
        )

        self.get_role_admin = GetRoleAdminMethod(
            web3_or_provider, contract_address, functions.getRoleAdmin, validator
        )

        self.get_role_member = GetRoleMemberMethod(
            web3_or_provider, contract_address, functions.getRoleMember, validator
        )

        self.get_role_member_count = GetRoleMemberCountMethod(
            web3_or_provider, contract_address, functions.getRoleMemberCount, validator
        )

        self.grant_role = GrantRoleMethod(
            web3_or_provider, contract_address, functions.grantRole, validator
        )

        self.has_role = HasRoleMethod(
            web3_or_provider, contract_address, functions.hasRole, validator
        )

        self.is_trusted_forwarder = IsTrustedForwarderMethod(
            web3_or_provider, contract_address, functions.isTrustedForwarder, validator
        )

        self.multicall = MulticallMethod(
            web3_or_provider, contract_address, functions.multicall, validator
        )

        self.remove = RemoveMethod(
            web3_or_provider, contract_address, functions.remove, validator
        )

        self.renounce_role = RenounceRoleMethod(
            web3_or_provider, contract_address, functions.renounceRole, validator
        )

        self.revoke_role = RevokeRoleMethod(
            web3_or_provider, contract_address, functions.revokeRole, validator
        )

        self.supports_interface = SupportsInterfaceMethod(
            web3_or_provider, contract_address, functions.supportsInterface, validator
        )

    def get_added_event(self, tx_hash: Union[HexBytes, bytes]) -> Tuple[AttributeDict]:
        """Get log entry for Added event.

        :param tx_hash: hash of transaction emitting Added event
        """
        tx_receipt = self._web3_eth.getTransactionReceipt(tx_hash)
        return (
            self._web3_eth.contract(
                address=to_checksum_address(self.contract_address), abi=TWRegistry.abi()
            )
            .events.Added()
            .processReceipt(tx_receipt)
        )

    def get_deleted_event(
        self, tx_hash: Union[HexBytes, bytes]
    ) -> Tuple[AttributeDict]:
        """Get log entry for Deleted event.

        :param tx_hash: hash of transaction emitting Deleted event
        """
        tx_receipt = self._web3_eth.getTransactionReceipt(tx_hash)
        return (
            self._web3_eth.contract(
                address=to_checksum_address(self.contract_address), abi=TWRegistry.abi()
            )
            .events.Deleted()
            .processReceipt(tx_receipt)
        )

    def get_role_admin_changed_event(
        self, tx_hash: Union[HexBytes, bytes]
    ) -> Tuple[AttributeDict]:
        """Get log entry for RoleAdminChanged event.

        :param tx_hash: hash of transaction emitting RoleAdminChanged event
        """
        tx_receipt = self._web3_eth.getTransactionReceipt(tx_hash)
        return (
            self._web3_eth.contract(
                address=to_checksum_address(self.contract_address), abi=TWRegistry.abi()
            )
            .events.RoleAdminChanged()
            .processReceipt(tx_receipt)
        )

    def get_role_granted_event(
        self, tx_hash: Union[HexBytes, bytes]
    ) -> Tuple[AttributeDict]:
        """Get log entry for RoleGranted event.

        :param tx_hash: hash of transaction emitting RoleGranted event
        """
        tx_receipt = self._web3_eth.getTransactionReceipt(tx_hash)
        return (
            self._web3_eth.contract(
                address=to_checksum_address(self.contract_address), abi=TWRegistry.abi()
            )
            .events.RoleGranted()
            .processReceipt(tx_receipt)
        )

    def get_role_revoked_event(
        self, tx_hash: Union[HexBytes, bytes]
    ) -> Tuple[AttributeDict]:
        """Get log entry for RoleRevoked event.

        :param tx_hash: hash of transaction emitting RoleRevoked event
        """
        tx_receipt = self._web3_eth.getTransactionReceipt(tx_hash)
        return (
            self._web3_eth.contract(
                address=to_checksum_address(self.contract_address), abi=TWRegistry.abi()
            )
            .events.RoleRevoked()
            .processReceipt(tx_receipt)
        )

    @staticmethod
    def abi():
        """Return the ABI to the underlying contract."""
        return json.loads(
            '[{"inputs":[{"internalType":"address","name":"_trustedForwarder","type":"address"}],"stateMutability":"nonpayable","type":"constructor"},{"anonymous":false,"inputs":[{"indexed":true,"internalType":"address","name":"deployer","type":"address"},{"indexed":true,"internalType":"address","name":"deployment","type":"address"}],"name":"Added","type":"event"},{"anonymous":false,"inputs":[{"indexed":true,"internalType":"address","name":"deployer","type":"address"},{"indexed":true,"internalType":"address","name":"deployment","type":"address"}],"name":"Deleted","type":"event"},{"anonymous":false,"inputs":[{"indexed":true,"internalType":"bytes32","name":"role","type":"bytes32"},{"indexed":true,"internalType":"bytes32","name":"previousAdminRole","type":"bytes32"},{"indexed":true,"internalType":"bytes32","name":"newAdminRole","type":"bytes32"}],"name":"RoleAdminChanged","type":"event"},{"anonymous":false,"inputs":[{"indexed":true,"internalType":"bytes32","name":"role","type":"bytes32"},{"indexed":true,"internalType":"address","name":"account","type":"address"},{"indexed":true,"internalType":"address","name":"sender","type":"address"}],"name":"RoleGranted","type":"event"},{"anonymous":false,"inputs":[{"indexed":true,"internalType":"bytes32","name":"role","type":"bytes32"},{"indexed":true,"internalType":"address","name":"account","type":"address"},{"indexed":true,"internalType":"address","name":"sender","type":"address"}],"name":"RoleRevoked","type":"event"},{"inputs":[],"name":"DEFAULT_ADMIN_ROLE","outputs":[{"internalType":"bytes32","name":"","type":"bytes32"}],"stateMutability":"view","type":"function"},{"inputs":[],"name":"OPERATOR_ROLE","outputs":[{"internalType":"bytes32","name":"","type":"bytes32"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"address","name":"_deployer","type":"address"},{"internalType":"address","name":"_deployment","type":"address"}],"name":"add","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"address","name":"_deployer","type":"address"}],"name":"count","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"address","name":"_deployer","type":"address"}],"name":"getAll","outputs":[{"internalType":"address[]","name":"","type":"address[]"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"}],"name":"getRoleAdmin","outputs":[{"internalType":"bytes32","name":"","type":"bytes32"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"},{"internalType":"uint256","name":"index","type":"uint256"}],"name":"getRoleMember","outputs":[{"internalType":"address","name":"","type":"address"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"}],"name":"getRoleMemberCount","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"},{"internalType":"address","name":"account","type":"address"}],"name":"grantRole","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"},{"internalType":"address","name":"account","type":"address"}],"name":"hasRole","outputs":[{"internalType":"bool","name":"","type":"bool"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"address","name":"forwarder","type":"address"}],"name":"isTrustedForwarder","outputs":[{"internalType":"bool","name":"","type":"bool"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"bytes[]","name":"data","type":"bytes[]"}],"name":"multicall","outputs":[{"internalType":"bytes[]","name":"results","type":"bytes[]"}],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"address","name":"_deployer","type":"address"},{"internalType":"address","name":"_deployment","type":"address"}],"name":"remove","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"},{"internalType":"address","name":"account","type":"address"}],"name":"renounceRole","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"},{"internalType":"address","name":"account","type":"address"}],"name":"revokeRole","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"bytes4","name":"interfaceId","type":"bytes4"}],"name":"supportsInterface","outputs":[{"internalType":"bool","name":"","type":"bool"}],"stateMutability":"view","type":"function"}]'  # noqa: E501 (line-too-long)
        )


# pylint: disable=too-many-lines



from enum import Enum
from typing import TypeVar, Union
from thirdweb.abi import (
    TWRegistry,
    TWFactory,
    TokenERC20,
    TokenERC721,
    TokenERC1155,
    Marketplace,
    IERC20,
    IERC721,
    IERC1155,
    DropERC721,
    DropERC1155,
    ThirdwebContract,
)
from thirdweb.types.settings.metadata import ContractMetadataSchema


TContractABI = TypeVar(
    "TContractABI",
    bound=Union[
        TokenERC721,
        TokenERC1155,
        TokenERC20,
        Marketplace,
        IERC20,
        IERC721,
        IERC1155,
        TWRegistry,
        TWFactory,
        DropERC721,
        DropERC1155,
        ThirdwebContract,
    ],
)

TERC721 = TypeVar("TERC721", bound=Union[TokenERC721, DropERC721])
TERC1155 = TypeVar("TERC1155", bound=Union[TokenERC1155, DropERC1155])

TPrimarySaleABI = TypeVar(
    "TPrimarySaleABI",
    bound=Union[TokenERC721, TokenERC1155, TokenERC20, DropERC721, DropERC1155],
)

TPlatformFeeABI = TypeVar(
    "TPlatformFeeABI",
    bound=Union[
        TokenERC721, TokenERC1155, TokenERC20, Marketplace, DropERC721, DropERC1155
    ],
)

TRoyaltyABI = TypeVar(
    "TRoyaltyABI",
    bound=Union[TokenERC721, TokenERC1155, DropERC721, DropERC1155],
)

TMetadataABI = TypeVar(
    "TMetadataABI",
    bound=Union[
        TokenERC721,
        TokenERC1155,
        TokenERC20,
        Marketplace,
        DropERC721,
        DropERC1155,
    ],
)

TContractSchema = TypeVar("TContractSchema", bound=ContractMetadataSchema)


class ContractType(Enum):
    NFT_COLLECTION = "nft-collection"
    EDITION = "edition"
    TOKEN = "token"
    MARKETPLACE = "marketplace"
    NFT_DROP = "nft-drop"
    EDITION_DROP = "edition-drop"
    CUSTOM = "custom"



"""Interface for interacting with an nft collection contract"""

from thirdweb.constants.role import Role
from thirdweb.core.classes.contract_events import ContractEvents
from thirdweb.core.classes.contract_metadata import ContractMetadata
from thirdweb.core.classes.contract_platform_fee import ContractPlatformFee
from thirdweb.core.classes.contract_roles import ContractRoles
from thirdweb.core.classes.contract_royalty import ContractRoyalty
from thirdweb.core.classes.contract_sales import ContractPrimarySale
from thirdweb.core.classes.contract_wrapper import ContractWrapper
from thirdweb.common.nft import upload_or_extract_uri, upload_or_extract_uris
from thirdweb.core.classes.erc_721 import ERC721
from thirdweb.core.classes.erc_721_signature_minting import ERC721SignatureMinting
from thirdweb.abi import TokenERC721

from thirdweb.core.classes.ipfs_storage import IpfsStorage
from thirdweb.types.contract import ContractType
from thirdweb.types.nft import NFTMetadataInput, NFTMetadataOwner
from eth_account.account import LocalAccount
from web3 import Web3

from thirdweb.types.sdk import SDKOptions
from typing import Final, Optional, List, Union

from thirdweb.types.settings.metadata import NFTCollectionContractMetadata
from thirdweb.types.tx import TxResultWithId


class NFTCollection(ERC721[TokenERC721]):
    """
    Create a collection of one-of-one NFTs.

    ```python
    from thirdweb import ThirdwebSDK
    from eth_account import Account

    # You can customize this to a supported network or your own RPC URL
    network = "mumbai"

    # This will create a random account to use for signing transactions
    signer = Account.create()

    sdk = ThirdwebSDK(network, signer)
    contract = sdk.get_nft_collection("{{contract_address}}")
    ```
    """

    _abi_type = TokenERC721

    contract_type: Final[ContractType] = ContractType.NFT_COLLECTION
    contract_roles: Final[List[Role]] = [Role.ADMIN, Role.MINTER, Role.TRANSFER]

    metadata: ContractMetadata[TokenERC721, NFTCollectionContractMetadata]
    roles: ContractRoles
    primary_sale: ContractPrimarySale[TokenERC721]
    platform_fee: ContractPlatformFee[TokenERC721]
    royalty: ContractRoyalty[TokenERC721]
    signature: ERC721SignatureMinting
    events: ContractEvents[TokenERC721]

    def __init__(
        self,
        provider: Web3,
        address: str,
        storage: IpfsStorage,
        signer: Optional[LocalAccount] = None,
        options: SDKOptions = SDKOptions(),
    ):
        abi = TokenERC721(provider, address)
        contract_wrapper = ContractWrapper(abi, provider, signer, options)
        super().__init__(contract_wrapper, storage)

        self.metadata = ContractMetadata(
            contract_wrapper, storage, NFTCollectionContractMetadata
        )
        self.roles = ContractRoles(contract_wrapper, self.contract_roles)
        self.primary_sale = ContractPrimarySale(contract_wrapper)
        self.platform_fee = ContractPlatformFee(contract_wrapper)
        self.royalty = ContractRoyalty(contract_wrapper, self.metadata)
        self.signature = ERC721SignatureMinting(
            contract_wrapper, self.roles, self._storage
        )
        self.events = ContractEvents(contract_wrapper)

    """
    WRITE FUNCTIONS
    """

    def mint(
        self, metadata: Union[NFTMetadataInput, str]
    ) -> TxResultWithId[NFTMetadataOwner]:
        """
        Mint a new NFT to the connected wallet

        :param metadata: metadata of the NFT to mint
        :returns: receipt, id, and metadata for the mint
        """

        return self.mint_to(self._contract_wrapper.get_signer_address(), metadata)

    def mint_to(
        self, to: str, metadata: Union[NFTMetadataInput, str]
    ) -> TxResultWithId[NFTMetadataOwner]:
        """
        Mint a new NFT to the specified wallet

        ```python
        from thirdweb.types.nft import NFTMetadataInput

        # Note that you can customize this metadata however you like
        metadata = NFTMetadataInput.from_json({
            "name": "Cool NFT",
            "description": "This is a cool NFT",
            "image": open("path/to/file.jpg", "rb"),
        })

        # You can pass in any address here to mint the NFT to
        tx = contract.mint_to("{{wallet_address}}", metadata)
        receipt = tx.receipt
        token_id = tx.id
        nft = tx.data()
        ```

        :param to: wallet address to mint the NFT to
        :param metadata: metadata of the NFT to mint
        :returns: receipt, id, and metadata for the mint
        """

        uri = upload_or_extract_uri(metadata, self._storage)
        receipt = self._contract_wrapper.send_transaction("mint_to", [to, uri])
        events = self._contract_wrapper.get_events("TokensMinted", receipt)

        if len(events) == 0:
            raise Exception("No TokensMinted event found")

        id = events[0].get("args").get("tokenIdMinted")  # type: ignore

        return TxResultWithId(receipt, id=id, data=lambda: self.get(id))

    def mint_batch(
        self, metadatas: List[Union[NFTMetadataInput, str]]
    ) -> List[TxResultWithId[NFTMetadataOwner]]:
        """
        Mint a batch of new NFTs to the connected wallet

        :param metadatas: list of metadata of the NFTs to mint
        :returns: receipts, ids, and metadatas for each mint
        """

        return self.mint_batch_to(
            self._contract_wrapper.get_signer_address(), metadatas
        )

    def mint_batch_to(
        self, to: str, metadatas: List[Union[NFTMetadataInput, str]]
    ) -> List[TxResultWithId[NFTMetadataOwner]]:
        """
        Mint a batch of new NFTs to the specified wallet

        ```python
        from thirdweb.types.nft import NFTMetadataInput

        # You can customize this metadata however you like
        metadatas = [
            NFTMetadataInput.from_json({
                "name": "Cool NFT",
                "description": "This is a cool NFT",
                "image": open("path/to/file.jpg", "rb"),
            }),
            NFTMetadataInput.from_json({
                "name": "Cooler NFT",
                "description": "This is a cooler NFT",
                "image": open("path/to/file.jpg", "rb"),
            }),
        ]

        # You can pass in any address here to mint the NFT to
        txs = contract.mint_batch_to("{{wallet_address}}", metadatas)
        receipt = txs[0].receipt
        first_token_id = txs[0].id
        first_nft = txs[0].data()
        ```

        :param to: wallet address to mint the NFTs to
        :param metadatas: list of metadata of the NFTs to mint
        :returns: receipts, ids, and metadatas for each mint
        """

        uris = upload_or_extract_uris(metadatas, self._storage)

        encoded = []
        interface = self._contract_wrapper.get_contract_interface()
        for uri in uris:
            encoded.append(interface.encodeABI("mintTo", [to, uri]))

        receipt = self._contract_wrapper.multi_call(encoded)
        events = self._contract_wrapper.get_events("TokensMinted", receipt)

        if len(events) == 0 or len(events) < len(metadatas):
            raise Exception("No TokensMinted event found, minting failed")

        results = []
        for event in events:
            id = event.get("args").get("tokenIdMinted")  # type: ignore
            results.append(TxResultWithId(receipt, id=id, data=lambda: self.get(id)))

        return results



from typing import List, Generic, Optional
from thirdweb.abi import TokenERC721
from thirdweb.common.error import NotFoundException
from thirdweb.common.nft import fetch_token_metadata
from thirdweb.constants.currency import ZERO_ADDRESS
from thirdweb.constants.role import Role, get_role_hash
from thirdweb.core.classes.contract_wrapper import ContractWrapper
from thirdweb.core.classes.base_contract import BaseContract
from thirdweb.core.classes.ipfs_storage import IpfsStorage
from thirdweb.types.contract import TERC721
from thirdweb.types.nft import NFTMetadata, NFTMetadataOwner, QueryAllParams
from web3.eth import TxReceipt


class ERC721(Generic[TERC721], BaseContract[TERC721]):
    _storage: IpfsStorage

    def __init__(
        self,
        contract_wrapper: ContractWrapper,
        storage: IpfsStorage,
    ):
        super().__init__(contract_wrapper)
        self._storage = storage

    """
    READ FUNCTIONS
    """

    def get(self, token_id: int) -> NFTMetadataOwner:
        """
        Get metadata for a token

        ```python
        nft = contract.get(0)
        print(nft)
        ```

        :param token_id: token ID of the token to get the metadata for
        :return: the metadata for the token and its owner
        """

        try:
            owner = self.owner_of(token_id)
        except:
            owner = ZERO_ADDRESS

        metadata = self._get_token_metadata(token_id)
        return NFTMetadataOwner(metadata, owner)

    def get_all(
        self, query_params: QueryAllParams = QueryAllParams()
    ) -> List[NFTMetadataOwner]:
        """
        Get the metadata of all tokens in the contract

        ```python
        nfts = contract.get_all()
        print(nfts)
        ```

        :param query_params: optionally define a QueryAllParams instance to narrow the metadata query to specific tokens
        :return: the metadata of all tokens in the contract
        """

        max_id = min(query_params.start + query_params.count, self.get_total_count())
        return [self.get(token_id) for token_id in range(query_params.start, max_id)]

    def get_total_count(self) -> int:
        """
        Get the total number of NFTs minted by this contract

        :return: the total number of NFTs minted by this contract
        """

        return self._contract_wrapper._contract_abi.next_token_id_to_mint.call()

    def get_owned(self, address: str = "") -> List[NFTMetadataOwner]:
        """
        Get the metadata of all tokens owned by a specific address

        ```python
        nfts = contract.get_owned("{{wallet_address}}")
        print(nfts)
        ```

        :param address: the address to get the metadata for
        :return: the metadata of all tokens owned by the address
        """

        token_ids = self.get_owned_token_ids(address)
        return [self.get(token_id) for token_id in token_ids]

    def get_owned_token_ids(self, address: str = "") -> List[int]:
        """
        Get the token IDs owned by a specific address

        :param address: the address to get the token IDs for
        :return: the token IDs owned by the address
        """

        owner = address if address else self._contract_wrapper.get_signer_address()
        balance = self._contract_wrapper._contract_abi.balance_of.call(owner)
        return [
            self._contract_wrapper._contract_abi.token_of_owner_by_index.call(owner, i)
            for i in range(balance)
        ]

    def owner_of(self, token_id: int) -> str:
        """
        Get the owner of a token

        :param token_id: the token ID of the token to get the owner of
        :return: the owner of the token
        """
        return self._contract_wrapper._contract_abi.owner_of.call(token_id)

    def total_supply(
        self,
    ) -> int:
        """
        Get the total number of tokens in the contract

        :return: the total number of tokens in the contract
        """
        return self._contract_wrapper._contract_abi.next_token_id_to_mint.call()

    def balance(
        self,
    ) -> int:
        """
        Get the token balance of the connected wallet

        :return: the token balance of the connected wallet
        """

        return self.balance_of(self._contract_wrapper.get_signer_address())

    def balance_of(self, address: str) -> int:
        """
        Get the token balance of a specific address

        ```python
        balance = contract.balance_of("{{wallet_address}}")
        print(balance)
        ```

        :param address: the address to get the token balance of
        """

        return self._contract_wrapper._contract_abi.balance_of.call(address)

    def is_transfer_restricted(
        self,
    ) -> bool:
        """
        Check if the contract is restricted to transfers only by admins

        :return: True if the contract is restricted to transfers only by admins, False otherwise
        """

        anyone_can_transfer = self._contract_wrapper._contract_abi.has_role.call(
            get_role_hash(Role.TRANSFER), ZERO_ADDRESS
        )

        return not anyone_can_transfer

    def is_approved(self, address: str, operator: str) -> bool:
        """
        Check whether an operator address is approved for all operations of a specific addresses assets

        :param address: the address whose assets are to be checked
        :param operator: the address of the operator to check
        :return: True if the operator is approved for all operations of the assets, False otherwise
        """

        return self._contract_wrapper._contract_abi.is_approved_for_all.call(
            address, operator
        )

    """
    WRITE FUNCTIONS
    """

    def transfer(self, to: str, token_id: int) -> TxReceipt:
        """
        Transfer a specified token from the connected wallet to a specified address.

        ```python
        to = "{{wallet_address}}"
        token_id = 0

        receipt = contract.transfer(to, token_id)
        ```

        :param to: wallet address to transfer the tokens to
        :param token_id: the specific token ID to transfer
        :returns: transaction receipt of the transfer
        """

        fr = self._contract_wrapper.get_signer_address()
        return self._contract_wrapper.send_transaction(
            "safe_transfer_from1", [fr, to, token_id]
        )

    def burn(self, token_id: int) -> TxReceipt:
        """
        Burn a specified token from the connected wallet.

        :param token_id: token ID of the token to burn
        :returns: transaction receipt of the burn
        """

        return self._contract_wrapper.send_transaction("burn", [token_id])

    def set_approval_for_all(self, operator: str, approved: bool) -> TxReceipt:
        """
        Set the approval of an operator for all operations of a specific address's assets

        :param operator: the address of the operator to set the approval for
        :param approved: the address whos assets the operator is approved to manage
        :returns: transaction receipt of the approval setting
        """

        return self._contract_wrapper.send_transaction(
            "set_approval_for_all", [operator, approved]
        )

    """
    INTERNAL FUNCTIONS
    """

    def _get_token_metadata(self, token_id: int) -> NFTMetadata:
        token_uri = self._contract_wrapper._contract_abi.token_uri.call(token_id)

        if not token_uri:
            raise NotFoundException(str(token_id))

        return fetch_token_metadata(token_id, token_uri, self._storage)



"""Interface for interacting with a marketplace contract"""

from time import time
from thirdweb.common.error import ListingNotFoundException
from thirdweb.constants.currency import ZERO_ADDRESS
from thirdweb.core.classes.contract_events import ContractEvents
from thirdweb.core.classes.contract_platform_fee import ContractPlatformFee
from thirdweb.types.marketplace import (
    AuctionListing,
    ContractListing,
    DirectListing,
    ListingType,
    MarketplaceFilter,
)
from thirdweb.types.settings.metadata import MarketplaceContractMetadata
from thirdweb.core.classes.contract_metadata import ContractMetadata
from thirdweb.core.classes.contract_roles import ContractRoles
from thirdweb.core.classes.contract_wrapper import ContractWrapper
from thirdweb.core.classes.base_contract import BaseContract
from thirdweb.core.classes.ipfs_storage import IpfsStorage
from thirdweb.core.classes.marketplace_auction import MarketplaceAuction
from thirdweb.core.classes.marketplace_direct import MarketplaceDirect
from thirdweb.types.contract import ContractType
from eth_account.account import LocalAccount
from thirdweb.types.sdk import SDKOptions
from thirdweb.constants.role import Role, get_role_hash
from typing import Any, Final, List, Optional, Union, cast
from thirdweb.abi import Marketplace as MarketplaceABI
from web3.eth import TxReceipt
from web3 import Web3


class Marketplace(BaseContract[MarketplaceABI]):
    """
    Create your own whitelabel marketplace that enables users to buy and sell any digital assets.

    ```python
    from thirdweb import ThirdwebSDK
    from eth_account import Account

    # You can customize this to a supported network or your own RPC URL
    network = "mumbai"

    # This will create a random account to use for signing transactions
    signer = Account.create()

    sdk = ThirdwebSDK(network, signer)
    contract = sdk.get_marketplace("{{contract_address}}")
    ```
    """

    _abi_type = MarketplaceABI
    _storage: IpfsStorage

    contract_type: Final[ContractType] = ContractType.MARKETPLACE
    contract_roles: Final[List[Role]] = [Role.ADMIN, Role.LISTER, Role.ASSET]

    metadata: ContractMetadata[MarketplaceABI, MarketplaceContractMetadata]
    roles: ContractRoles
    platform_fee: ContractPlatformFee[MarketplaceABI]
    direct: MarketplaceDirect
    auction: MarketplaceAuction
    events: ContractEvents[MarketplaceABI]

    def __init__(
        self,
        provider: Web3,
        address: str,
        storage: IpfsStorage,
        signer: Optional[LocalAccount] = None,
        options: SDKOptions = SDKOptions(),
    ):
        abi = MarketplaceABI(provider, address)
        contract_wrapper = ContractWrapper(abi, provider, signer, options)
        super().__init__(contract_wrapper)

        self._storage = storage
        self.metadata = ContractMetadata(
            contract_wrapper, storage, MarketplaceContractMetadata
        )
        self.platform_fee = ContractPlatformFee(contract_wrapper)
        self.direct = MarketplaceDirect(contract_wrapper, storage)
        self.auction = MarketplaceAuction(contract_wrapper, storage)
        self.events = ContractEvents(contract_wrapper)

    """
    READ FUNCTIONS
    """

    def get_listing(self, listing_id: int) -> Union[DirectListing, AuctionListing]:
        """
        Get a listing from the marketplace by ID

        :param listing_id: ID of the listing to get
        :return: Listing object
        """
        listing_args = self._contract_wrapper._contract_abi.listings.call(listing_id)
        listing = ContractListing(*listing_args)

        if listing.asset_contract == ZERO_ADDRESS:
            raise ListingNotFoundException(listing_id)

        if ListingType(listing.listing_type) == ListingType.AUCTION:
            return self.auction._map_listing(listing)
        if ListingType(listing.listing_type) == ListingType.DIRECT:
            return self.direct._map_listing(listing)

        raise Exception(f"Unkown listing type {listing.listing_type}")

    def get_active_listings(self) -> List[Union[DirectListing, AuctionListing]]:
        """
        Get all the currently active listings from the marketplace.

        ```python
        listings = contract.get_active_listings()
        price_of_first = listings[0].price
        ```

        :return: List of listings
        """
        raw_listings = self._get_all_listings_no_filter()

        listings: List[Union[DirectListing, AuctionListing]] = []
        for listing in raw_listings:
            if ListingType(listing.type) == ListingType.AUCTION and cast(
                AuctionListing, listing
            ).end_time_in_epoch_seconds > int(time()):
                listings.append(listing)
            elif (
                ListingType(listing.type) == ListingType.DIRECT and listing.quantity > 0
            ):
                listings.append(listing)

        return listings

    def get_all_listings(
        self, filter: MarketplaceFilter = None
    ) -> List[Union[DirectListing, AuctionListing]]:
        """
        Get all the listings that have ever been made on this marketplace.

        ```python
        listings = contract.get_all_listings()
        price_of_first = listings[0].price
        ```

        :param filter: Filter to apply to the listings
        :return: List of listings
        """
        raw_listings = self._get_all_listings_no_filter()

        if filter is not None:
            if filter.seller:
                raw_listings = [
                    listing
                    for listing in raw_listings
                    if listing.seller_address == filter.seller
                ]
            if filter.token_contract:
                if filter.token_id:
                    raw_listings = [
                        listing
                        for listing in raw_listings
                        if listing.asset_contract_address == filter.token_contract
                        and listing.token_id == filter.token_id
                    ]
                else:
                    raw_listings = [
                        listing
                        for listing in raw_listings
                        if listing.asset_contract_address == filter.token_contract
                    ]

            max_id = min(filter.start + filter.count, self.get_total_count())
            raw_listings = [raw_listings[i] for i in range(filter.start, max_id)]

        return [listing for listing in raw_listings if listing is not None]

    get_all = get_all_listings

    def get_total_count(self) -> int:
        """
        Get the total number of listings on this marketplace.

        :return: Total number of listings
        """
        return self._contract_wrapper._contract_abi.total_listings.call()

    def is_restricted_to_lister_role_only(self) -> bool:
        """
        Check whether only wallets with the lister role can make listings.

        :return: True if only lister wallets can make listings
        """
        anyone_can_list = self._contract_wrapper._contract_abi.has_role.call(
            get_role_hash(Role.LISTER), ZERO_ADDRESS
        )
        return not anyone_can_list

    def get_bid_buffer_bps(self) -> int:
        """
        Get the bid buffer basis points for this marketplace.

        :return: Bid buffer basis points
        """
        return self._contract_wrapper._contract_abi.bid_buffer_bps.call()

    def get_time_buffer_in_seconds(self) -> int:
        """
        Get the time buffer for this marketplace in seconds

        :return: Time buffer in seconds
        """
        return self._contract_wrapper._contract_abi.time_buffer.call()

    """
    WRITE FUNCTIONS
    """

    def buyout_listing(
        self,
        listing_id: int,
        quantity_desired: Optional[int] = None,
        receiver: Optional[str] = None,
    ) -> TxReceipt:
        """
        Buyout a listing by listing ID

        ```python
        listing_id = 0
        quantity_desired = 1

        contract.buyout_listing(listing_id, quantity_desired)
        ```

        :param listing_id: ID of the listing to buyout
        :param quantity_desired: Quantity to buyout
        :param receiver: Address to send the asset to
        :return: Transaction receipt of buyout
        """
        raw_listing = self._contract_wrapper._contract_abi.listings.call(listing_id)
        listing = ContractListing(*raw_listing)

        if listing.listing_id != listing_id:
            raise ListingNotFoundException(listing_id)

        if ListingType(listing.listing_type) == ListingType.DIRECT:
            if quantity_desired is None:
                raise Exception("quantity_desired is required for direct listings")

            return self.direct.buyout_listing(listing_id, quantity_desired, receiver)
        elif ListingType(listing.listing_type) == ListingType.AUCTION:
            return self.auction.buyout_listing(listing_id)

        raise Exception(f"Unkown listing type {listing.listing_type}")

    def set_bid_buffer_bps(self, buffer_bps: int) -> TxReceipt:
        """
        Set the bid buffer basis points for this marketplace.

        ```python
        buffer_bps = 500
        contract.set_bid_buffer_bps(buffer_bps)
        ```

        :param buffer_bps: Bid buffer basis points
        :return: Transaction receipt
        """
        self.roles.verify([Role.ADMIN], self._contract_wrapper.get_signer_address())

        time_buffer = self.get_time_buffer_in_seconds()
        return self._contract_wrapper.send_transaction(
            "set_auction_buffers", [time_buffer, buffer_bps]
        )

    def set_time_buffer_in_seconds(self, buffer_in_seconds: int) -> TxReceipt:
        """
        Set the time buffer of the marketplace.

        ```python
        buffer_in_seconds = 60
        contract.set_time_buffer_in_seconds(buffer_in_seconds)
        ```

        :param buffer_in_seconds: Time buffer in seconds
        :return: Transaction receipt
        """
        self.roles.verify([Role.ADMIN], self._contract_wrapper.get_signer_address())

        bid_buffer_bps = self.get_bid_buffer_bps()
        return self._contract_wrapper.send_transaction(
            "set_auction_buffers", [buffer_in_seconds, bid_buffer_bps]
        )

    def allow_listing_from_specific_asset_only(
        self, contract_address: str
    ) -> TxReceipt:
        """
        Restrict marketplace so only specific asset can be listed.

        :param contract_address: Address of the asset contract
        """
        encoded = []
        interface = self._contract_wrapper.get_contract_interface()
        members = self.roles.get(Role.ASSET)

        if ZERO_ADDRESS in members:
            encoded.append(
                interface.encodeABI(
                    "revoke_role", [get_role_hash(Role.ASSET), ZERO_ADDRESS]
                )
            )

        encoded.append(
            interface.encodeABI(
                "grant_role", [get_role_hash(Role.ASSET), contract_address]
            )
        )

        return self._contract_wrapper.multi_call(encoded)

    def allow_listing_from_any_asset(self) -> TxReceipt:
        """
        Allow asset to be listed on the marketplace.

        :return: Transaction receipt
        """

        encoded = []
        interface = self._contract_wrapper.get_contract_interface()
        members = self.roles.get(Role.ASSET)

        for member in members:
            encoded.append(
                interface.encodeABI("revoke_role", [get_role_hash(Role.ASSET), member])
            )

        encoded.append(
            interface.encodeABI("grant_role", [get_role_hash(Role.ASSET), ZERO_ADDRESS])
        )

        return self._contract_wrapper.multi_call(encoded)

    """
    INTERNAL FUNCTIONS
    """

    def _get_all_listings_no_filter(
        self,
    ) -> List[Union[DirectListing, AuctionListing]]:
        total_listings = self._contract_wrapper._contract_abi.total_listings.call()

        listings = []
        for i in range(total_listings):
            try:
                listing = self.get_listing(i)

                if ListingType(listing.type) == ListingType.AUCTION:
                    listings.append(listing)
                else:
                    valid = self.direct._is_still_valid_listing(cast(Any, listing))

                    if valid:
                        listings.append(listing)
            except:
                pass

        return [listing for listing in listings if listing is not None]



from typing import Any, Generic, Tuple, List, Optional, cast
from eth_typing import Address

from web3 import Web3
from web3.datastructures import AttributeDict
from web3.contract import Contract
from thirdweb.common.error import NoSignerException
from web3._utils.events import EventLogErrorFlags
from thirdweb.common.sign import EIP712Domain, sign_typed_data_internal
from thirdweb.constants.events import EventStatus, EventType

from thirdweb.core.classes.provider_handler import ProviderHandler
from web3.eth import TxReceipt
from eth_account.account import LocalAccount
from zero_ex.contract_wrappers.tx_params import TxParams
from thirdweb.types.contract import TContractABI
from thirdweb.types.events import SignatureEvent, TxEvent

from thirdweb.types.sdk import SDKOptions


class ContractWrapper(Generic[TContractABI], ProviderHandler):
    """
    The contract wrapper wraps an instance of a specific thirdweb contract ABI
    and exposed functions for interacting with the contract.
    """

    _contract_abi: TContractABI

    def __init__(
        self,
        contract_abi: TContractABI,
        provider: Web3,
        signer: Optional[LocalAccount] = None,
        options: SDKOptions = SDKOptions(),
    ):
        """
        Initializes the contract wrapper.

        :param contract_abi: ABI of the thirdweb contract to use
        :param provider: web3 provider instance to use
        :param signer: optional account to use for signing transactions
        """

        super().__init__(provider, signer, options)
        self._contract_abi = contract_abi

    def get_chain_id(self) -> int:
        """
        Get the chain ID of the active provider

        :returns: chain ID of the active provider
        """

        provider = self.get_provider()
        return provider.eth.chain_id

    def get_signer_address(self) -> str:
        """
        Get the address of the active signer

        :returns: address of the active signer
        """

        signer = self.get_signer()

        if signer is None:
            raise NoSignerException

        return signer.address

    def get_contract_interface(self) -> Contract:
        """
        Get the contract interface of the contract wrapper.

        :returns: contract interface of the contract wrapper
        """

        return self.get_provider().eth.contract(
            address=cast(Address, self._contract_abi.contract_address),
            abi=self._contract_abi.abi(),
        )

    def get_events(self, event: str, receipt: TxReceipt) -> Tuple[AttributeDict]:
        """
        Get the events from a transaction receipt.

        :param event: name of the event to get
        :param receipt: transaction receipt to get the events from
        """

        return (
            self.get_contract_interface()
            .events[event]()
            .processReceipt(receipt, errors=EventLogErrorFlags.Discard)
        )

    def send_transaction(self, fn: str, args: List[Any]) -> TxReceipt:
        """
        Send and execute a transaction and return the receipt.

        :param fn: name of the function you want to call on the contract
        :param args: list of arguments to pass to the function
        """

        provider = self.get_provider()
        signer = self.get_signer()

        if signer is None:
            raise NoSignerException

        nonce = provider.eth.get_transaction_count(signer.address)
        tx = getattr(self._contract_abi, fn).build_transaction(
            *args, tx_params=TxParams(gas_price=provider.eth.gas_price)
        )
        tx["nonce"] = nonce

        signed_tx = signer.sign_transaction(tx)
        tx_hash = provider.eth.send_raw_transaction(signed_tx.rawTransaction)

        self.emit_transaction_event(EventStatus.SUBMITTED, tx_hash.hex())

        receipt = provider.eth.wait_for_transaction_receipt(tx_hash)

        self.emit_transaction_event(EventStatus.COMPLETED, tx_hash.hex())

        return receipt

    def multi_call(self, encoded: List[str]) -> TxReceipt:
        """
        Execute a multicall and return the result.

        :param encoded: list of encoded function calls to execute
        """

        return self.send_transaction("multicall", [encoded])

    def emit_transaction_event(self, status: EventStatus, tx_hash: str):
        self.emit(EventType.TRANSACTION, TxEvent(status, tx_hash))  # type: ignore

    def sign_typed_data(
        self, signer: LocalAccount, domain: EIP712Domain, types: Any, message: Any
    ) -> bytes:
        self.emit(
            EventType.SIGNATURE,  # type: ignore
            SignatureEvent(EventStatus.SUBMITTED, message, ""),
        )

        signature = sign_typed_data_internal(
            self.get_provider(), signer, domain, types, message
        )

        self.emit(
            EventType.SIGNATURE,  # type: ignore
            SignatureEvent(EventStatus.COMPLETED, message, signature),
        )

        return signature



from eth_utils import encode_hex
from hexbytes import HexBytes
from web3 import Web3
from typing import Dict
from enum import Enum

from thirdweb.constants.addresses import DEFAULT_MERKLE_ROOT


class Role(Enum):
    ADMIN = "admin"
    TRANSFER = "transfer"
    MINTER = "minter"
    PAUSER = "pauser"
    EDITOR = "editor"
    LISTER = "lister"
    ASSET = "asset"


ROLE_MAP: Dict[Role, str] = {
    Role.ADMIN: "",
    Role.TRANSFER: "TRANSFER_ROLE",
    Role.MINTER: "MINTER_ROLE",
    Role.PAUSER: "PAUSER_ROLE",
    Role.EDITOR: "EDITOR_ROLE",
    Role.LISTER: "LISTER_ROLE",
    Role.ASSET: "ASSET_ROLE",
}

ALL_ROLES = [
    Role.ADMIN,
    Role.TRANSFER,
    Role.MINTER,
    Role.PAUSER,
    Role.EDITOR,
    Role.LISTER,
    Role.ASSET,
]


def get_role_hash(role: Role) -> HexBytes:
    if role == Role.ADMIN:
        return HexBytes(DEFAULT_MERKLE_ROOT)
    return Web3.keccak(text=ROLE_MAP[role])



from typing import Optional, Union, cast
from thirdweb.abi import IERC165, IERC721, IERC1155
from thirdweb.common.currency import fetch_currency_value
from thirdweb.constants.contract import INTERFACE_ID_IERC1155, INTERFACE_ID_IERC721
from eth_account.account import LocalAccount
from web3 import Web3
from thirdweb.core.classes.contract_wrapper import ContractWrapper
from thirdweb.types.currency import Price, PriceWei

from thirdweb.types.marketplace import (
    ContractOffer,
    NewAuctionListing,
    NewDirectListing,
    Offer,
)

MAX_BPS = 10000


def is_token_approved_for_marketplace(
    provider: Web3,
    marketplace_address: str,
    asset_contract: str,
    token_id: int,
    fr: str,
) -> bool:
    try:
        erc165 = IERC165(provider, asset_contract)
        is_erc721 = erc165.supports_interface.call(INTERFACE_ID_IERC721)
        is_erc1155 = erc165.supports_interface.call(INTERFACE_ID_IERC1155)

        if is_erc721:
            ierc721 = IERC721(provider, asset_contract)
            approved = ierc721.is_approved_for_all.call(fr, marketplace_address)

            if approved:
                return True

            return (
                ierc721.get_approved.call(token_id).lower()
                == marketplace_address.lower()
            )
        elif is_erc1155:
            ierc1155 = IERC1155(provider, asset_contract)
            return ierc1155.is_approved_for_all.call(fr, marketplace_address)
        else:
            print("Contract does not implement ERC1155 or ERC721")
            return False
    except:
        print("Failed to check if token is approved")
        return False


def handle_token_approval(
    provider: Web3,
    signer: Optional[LocalAccount],
    marketplace_address: str,
    asset_contract: str,
    token_id: int,
    fr: str,
):
    erc165 = IERC165(provider, asset_contract)
    is_erc721 = erc165.supports_interface.call(INTERFACE_ID_IERC721)
    is_erc1155 = erc165.supports_interface.call(INTERFACE_ID_IERC1155)

    if is_erc721:
        ierc721_abi = IERC721(provider, asset_contract)
        ierc721 = ContractWrapper[IERC721](ierc721_abi, provider, signer)
        approved = ierc721._contract_abi.is_approved_for_all.call(
            fr, marketplace_address
        )

        if not approved:
            is_token_approved = (
                ierc721._contract_abi.get_approved.call(token_id).lower()
                == marketplace_address.lower()
            )

            if not is_token_approved:
                ierc721.send_transaction(
                    "set_approval_for_all", [marketplace_address, True]
                )
    elif is_erc1155:
        ierc1155_abi = IERC1155(provider, asset_contract)
        ierc1155 = ContractWrapper[IERC1155](ierc1155_abi, provider, signer)
        approved = ierc1155._contract_abi.is_approved_for_all.call(
            fr, marketplace_address
        )

        if not approved:
            ierc1155.send_transaction(
                "set_approval_for_all", [marketplace_address, True]
            )
    else:
        raise Exception("Contract does not implement ERC1155 or ERC721")


def validate_new_listing_param(param: Union[NewDirectListing, NewAuctionListing]):
    if param.asset_contract_address == None:
        raise Exception("Asset contract address is required")
    if param.buyout_price_per_token == None:
        raise Exception("Buyout price per token is required")
    if param.listing_duration_in_seconds == None:
        raise Exception("Listing duration is required")
    if param.start_time_in_seconds == None:
        raise Exception("Start time is required")
    if param.token_id == None:
        raise Exception("Token id is required")
    if param.quantity == None:
        raise Exception("Quantity is required")

    if param.type == "NewAuctionListing":
        if cast(NewAuctionListing, param).reserve_price_per_token == None:
            raise Exception("Reserve price per token is required")


def map_offer(provider: Web3, listing_id: int, offer: ContractOffer) -> Offer:
    return Offer(
        quantity_desired=offer.quantity_wanted,
        price_per_token=offer.price_per_token,
        buyer_address=offer.offeror,
        currency_contract_address=offer.currency,
        currency_value=fetch_currency_value(
            provider, offer.currency, offer.quantity_wanted * offer.price_per_token
        ),
        listing_id=listing_id,
    )


def is_winning_bid(
    winning_price: PriceWei, new_bid_price: PriceWei, bid_buffer: int
) -> bool:
    buffer = ((new_bid_price - winning_price) * MAX_BPS) / winning_price
    return buffer >= bid_buffer
